"""Verify implementation of RPC file helpers."""

import collections.abc
import contextlib
import datetime
import pathlib
import typing

import pytest

from shv import SHVType
from shv.rpcapi import SHVBase
from shv.rpcapi.client import SHVClient
from shv.rpcdef.file import FileProviderRW, RpcAccess, RpcDir, RpcFile, RpcFileStat


class FilesProvider(SHVClient):
    APP_NAME = "testdev"

    def __init__(
        self,
        *args: typing.Any,  # noqa ANNN401
        path: pathlib.Path,
        **kwargs: typing.Any,  # noqa ANNN401
    ) -> None:
        self.path = path
        super().__init__(*args, **kwargs)  # notype

    def _ls(self, path: str) -> collections.abc.Iterator[str]:
        yield from super()._ls(path)
        if not path:
            yield "files"
        elif path == "files" or path.startswith("files/"):
            pth = self.path / path[6:]
            if pth.is_dir():
                yield from sorted(spth.name for spth in pth.iterdir())

    def _dir(self, path: str) -> collections.abc.Iterator[RpcDir]:
        yield from super()._dir(path)
        if path.startswith("files/"):
            pth = self.path / path[6:]
            if pth.is_file():
                yield from FileProviderRW.dir()

    async def _method_call(self, request: SHVBase.Request) -> SHVType:
        if request.path.startswith("files/"):
            pth = self.path / request.path[6:]
            if pth.is_file():
                with contextlib.suppress(NotImplementedError):
                    # TODO use also other providers
                    return await FileProviderRW.method_call(pth, request)
        return await super()._method_call(request)


@pytest.fixture(name="device")
async def fixture_device(shvbroker, url_test_device, tmp_path):
    """Run FilesProvider and provide instance to access it."""
    device = await FilesProvider.connect(url_test_device, path=tmp_path)
    yield device
    await device.disconnect()


@pytest.fixture(name="seed_tmp")
def fixture_seed_tmp(tmp_path):
    """Create some dummy files in tmp_path."""
    (tmp_path / "bin").write_bytes(b"abc")
    (tmp_path / "lorem.txt").write_text(
        "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua."
    )
    (tmp_path / "sub" / "subsub").mkdir(parents=True)


@pytest.mark.parametrize(
    "path,res",
    (
        ("test/device", [".app", "files"]),
        ("test/device/files", ["bin", "lorem.txt", "sub"]),
        ("test/device/files/bin", []),
        ("test/device/files/lorem.txt", []),
        ("test/device/files/sub", ["subsub"]),
        ("test/device/files/sub/subsub", []),
    ),
)
async def test_ls(client, device, seed_tmp, path, res):
    assert await client.ls(path) == res


@pytest.mark.parametrize(
    "path,res",
    (
        (
            "test/device/files/lorem.txt",
            [
                RpcDir.stddir(),
                RpcDir.stdls(),
                RpcDir.getter("stat", "n", "!stat", RpcAccess.READ),
                RpcDir.getter("size", "n", "i(0,)", RpcAccess.READ),
                RpcDir(
                    "crc",
                    param="[i(0,):offset,i(0,)|n:size]|n",
                    result="u(>32)",
                    access=RpcAccess.READ,
                ),
                RpcDir(
                    "sha1",
                    param="[i(0,):offset,i(0,)|n:size]|n",
                    result="b(20)",
                    access=RpcAccess.READ,
                ),
                RpcDir(
                    "read",
                    RpcDir.Flag.LARGE_RESULT_HINT,
                    "[i(0,):offset,i(0,)size]",
                    "b",
                    RpcAccess.READ,
                ),
                RpcDir(
                    "write",
                    param="[i(0,):offset,b:data]",
                    result="n",
                    access=RpcAccess.WRITE,
                ),
                RpcDir("truncate", param="i(0,)", result="n", access=RpcAccess.WRITE),
                RpcDir("append", param="b", result="n", access=RpcAccess.WRITE),
            ],
        ),
    ),
)
async def test_dir(client, device, seed_tmp, path, res):
    assert await client.dir(path) == res


@pytest.fixture(name="file_bin")
def fixture_file_bin(client, device, seed_tmp):
    return RpcFile(client, "test/device/files/bin", RpcFile.Flag(0))


def test_file_bin_path(file_bin):
    assert file_bin.path == "test/device/files/bin"


async def test_file_bin_readable(file_bin):
    assert await file_bin.readable()


async def test_file_bin_writable(file_bin):
    assert await file_bin.writable()


async def test_file_bin_resizable(file_bin):
    assert await file_bin.resizable()


async def test_file_bin_stat(file_bin, tmp_path):
    stat = (tmp_path / "bin").stat()
    access = datetime.datetime.fromtimestamp(stat.st_atime, tz=datetime.UTC)
    access = access.replace(microsecond=(access.microsecond // 1000) * 1000)
    mod = datetime.datetime.fromtimestamp(stat.st_mtime, tz=datetime.UTC)
    mod = access.replace(microsecond=(mod.microsecond // 1000) * 1000)
    assert await file_bin.stat() == RpcFileStat(
        0, stat.st_size, stat.st_blksize, access, mod
    )


async def test_file_bin_size(client, device, seed_tmp):
    assert await client.call("test/device/files/bin", "size") == 3


@pytest.mark.parametrize(
    "offset,size,res",
    (
        (0, None, 891568578),
        (1, None, 3265866552),
        (0, 1, 3904355907),
    ),
)
async def test_file_bin_crc32(file_bin, offset, size, res):
    assert await file_bin.crc32(offset, size) == res


@pytest.mark.parametrize(
    "offset,size,res",
    (
        (0, None, b"\xa9\x99>6G\x06\x81j\xba>%qxP\xc2l\x9c\xd0\xd8\x9d"),
        (1, None, b"[%\x05\x03\x9a\xc5\xaf\x9e\x19\x7f]\xad\x04\x119\x06\xa9\xcf\x9a*"),
        (0, 1, b"\x86\xf7\xe47\xfa\xa5\xa7\xfc\xe1]\x1d\xdc\xb9\xea\xea\xea7vg\xb8"),
    ),
)
async def test_file_bin_sha1(file_bin, offset, size, res):
    assert await file_bin.sha1(offset, size) == res


async def test_file_bin_read(file_bin):
    assert file_bin.offset == 0
    assert await file_bin.read() == b"abc"
    assert file_bin.offset == 3


async def test_file_bin_read_short(file_bin):
    assert file_bin.offset == 0
    assert await file_bin.read(2) == b"ab"
    assert file_bin.offset == 2
    assert await file_bin.read(2) == b"c"
    assert file_bin.offset == 3


async def test_file_bin_write(file_bin, tmp_path):
    file_bin.offset = 3
    await file_bin.write(b"def")
    assert file_bin.offset == 6
    with (tmp_path / "bin").open("rb") as file:
        assert file.read() == b"abcdef"


async def test_file_bin_flush(file_bin, tmp_path):
    # This does nothing so just call it that it won't fail
    await file_bin.flush()


async def test_file_bin_truncate(file_bin, tmp_path):
    await file_bin.truncate()
    with (tmp_path / "bin").open("rb") as file:
        assert file.read() == b""


@pytest.fixture(name="file_append")
def fixture_file_append(client, device, seed_tmp):
    return RpcFile(client, "test/device/files/bin", RpcFile.Flag.APPEND)


async def test_file_bin_append(file_append, tmp_path):
    file_append.offset = 0
    await file_append.write(b"def")
    file_append.offset = 0
    with (tmp_path / "bin").open("rb") as file:
        assert file.read() == b"abcdef"


@pytest.fixture(name="file_buffered")
def fixture_file_buffered(client, device, seed_tmp):
    return RpcFile(client, "test/device/files/bin", RpcFile.Flag.BUFFERED)


async def test_file_buffered_read(file_buffered, tmp_path):
    assert file_buffered.offset == 0
    assert await file_buffered.read(1) == b"a"
    assert file_buffered.offset == 1
    assert await file_buffered.read(1) == b"b"
    assert file_buffered.offset == 2
    assert await file_buffered.read(2) == b"c"
    assert file_buffered.offset == 3
    assert await file_buffered.read(2) == b""
    assert file_buffered.offset == 3
    with (tmp_path / "bin").open("ab") as file:
        file.write(b"def")
    assert await file_buffered.read(2) == b"de"
    assert file_buffered.offset == 5


async def test_file_buffered_write(file_buffered, tmp_path):
    file_buffered.offset = await file_buffered.size()
    await file_buffered.write(b"de")
    await file_buffered.write(b"f")
    with (tmp_path / "bin").open("rb") as file:
        assert file.read() == b"abc"
    await file_buffered.flush()
    with (tmp_path / "bin").open("rb") as file:
        assert file.read() == b"abcdef"

    await file_buffered.write(b"gh")
    file_buffered.offset = 1
    await file_buffered.write(b"d")
    with (tmp_path / "bin").open("rb") as file:
        assert file.read() == b"abcdefgh"



"""Setuptools hook used to replace __version__.py file."""

import pathlib

from setuptools.command.build_py import build_py as _build_py


class BuildPy(_build_py):
    """Custom package build step."""

    def run(self) -> None:  # noqa: D102
        package = pathlib.Path(__file__).parent.name
        buildlib = pathlib.Path(self.build_lib) / package
        buildlib.mkdir(parents=True, exist_ok=True)
        version = self.distribution.metadata.version
        (buildlib / "__version__.py").write_text(f'VERSION = "{version}"\n')
        super().run()
        (buildlib / "__setuptools__.py").unlink()



"""The SHV RPC type tuple."""

from __future__ import annotations

import collections.abc
import typing

from .. import SHVType
from .base import RpcType


class RpcTypeTuple(RpcType, collections.abc.Sequence[tuple[RpcType, str]]):
    """The Tuple type representation."""

    def __init__(self, *args: tuple[RpcType, str]) -> None:
        self._items = args

    def __eq__(self, other: object) -> bool:
        return isinstance(other, RpcTypeTuple) and self._items == other._items

    def __str__(self) -> str:
        return f"[{','.join(f'{tp}:{key}' for tp, key in self._items)}]"

    @typing.overload
    def __getitem__(self, i: int, /) -> tuple[RpcType, str]: ...

    @typing.overload
    def __getitem__(
        self, i: slice, /
    ) -> collections.abc.Sequence[tuple[RpcType, str]]: ...

    def __getitem__(
        self, i: int | slice
    ) -> tuple[RpcType, str] | collections.abc.Sequence[tuple[RpcType, str]]:
        return self._items[i]

    def __len__(self) -> int:
        return len(self._items)

    def validate(self, value: SHVType) -> typing.TypeGuard[collections.abc.Sequence]:  # noqa: D102
        if not isinstance(value, collections.abc.Sequence):
            return False
        if len(value) > len(self._items):
            return False
        for i, item in enumerate(self._items):
            if not item[0].validate(value[i] if len(value) > i else None):
                return False
        return True



"""Check the behavior of the SHVBase."""

import asyncio
import collections.abc
import logging

import pytest

from shv import SHV_VERSION, SHVType
from shv.rpcapi import SHVBase
from shv.rpcdef import (
    RpcDir,
    RpcInvalidParamError,
    RpcRequestInvalidError,
    RpcTryAgainLaterError,
    RpcUserIDRequiredError,
)
from shv.rpcmessage import RpcMessage
from shv.rpctransport import RpcClientPipe

logger = logging.getLogger(__name__)


class App(SHVBase):
    """The SHV application used in these tests."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.event = asyncio.Event()
        self.cancelled = False

    async def _method_call(self, request: SHVBase.Request) -> SHVType:
        match request.path, request.method:
            case "test", "delay":
                for i in range(4):
                    request.progress = 1.0 / 4 * i
                    await asyncio.sleep(0.01)
                return None
            case "test", "delay-event":
                try:
                    await self.event.wait()
                    request.progress = 0.42
                    self.event.clear()
                    await self.event.wait()
                    return None
                except asyncio.CancelledError:
                    self.cancelled = True
                    raise
        return await super()._method_call(request)

    def _ls(self, path: str) -> collections.abc.Iterator[str]:
        if not path:
            yield "test"
        yield from super()._ls(path)

    def _dir(self, path: str) -> collections.abc.Iterator[RpcDir]:
        if path == "test":
            yield RpcDir("delay")
            yield RpcDir("delay-event")
        yield from super()._dir(path)


@pytest.fixture(name="con")
async def fixture_con():
    """Pair of raw client and SHVBase."""
    client1, client2 = await RpcClientPipe.open_pair()
    shv = App(client2, peer_shv_version=SHV_VERSION)
    yield shv, client1
    client1.disconnect()
    await shv.disconnect()
    await client1.wait_disconnect()


async def test_req_ping(con):
    msg = RpcMessage.request(".app", "ping")
    await con[1].send(msg)
    assert await con[1].receive() == msg.make_response()


async def test_delay(con):
    """Check that we report progress on our own."""
    msg = RpcMessage.request("test", "delay")
    await con[1].send(msg)
    assert await con[1].receive() == msg.make_response_delay(0.0)
    assert await con[1].receive() == msg.make_response_delay(0.25)
    assert await con[1].receive() == msg.make_response_delay(0.5)
    assert await con[1].receive() == msg.make_response_delay(0.75)
    assert await con[1].receive() == msg.make_response()


async def test_delay_event(con):
    """Check that we can query the delayed task."""
    msg = RpcMessage.request("test", "delay-event")
    await con[1].send(msg)
    await con[1].send(msg.make_abort(False))
    assert await con[1].receive() == msg.make_response_delay(0.0)
    con[0].event.set()
    assert await con[1].receive() == msg.make_response_delay(0.42)
    await con[1].send(msg.make_abort(False))
    assert await con[1].receive() == msg.make_response_delay(0.42)
    con[0].event.set()
    assert await con[1].receive() == msg.make_response()


async def test_delay_event_abort(con):
    """Check that we can query the delayed task."""
    msg = RpcMessage.request("test", "delay-event")
    await con[1].send(msg)
    con[0].event.set()
    assert await con[1].receive() == msg.make_response_delay(0.42)
    assert not con[0].cancelled
    await con[1].send(msg.make_abort(True))
    assert await con[1].receive() == msg.make_response(
        RpcRequestInvalidError("Request cancelled")
    )
    assert con[0].cancelled


async def test_call_ping(con):
    """Check that we correctly simple call sequence."""
    task = asyncio.create_task(con[0].ping())
    msg = await con[1].receive()
    assert msg == RpcMessage.request(
        ".app", "ping", None, RpcMessage.next_request_id() - 1
    )
    await con[1].send(msg.make_response())
    assert await task is None


async def test_call_error(con):
    """Check that we propagate error."""
    task = asyncio.create_task(con[0].call("test", "error"))
    msg = await con[1].receive()
    assert msg == RpcMessage.request(
        "test", "error", None, RpcMessage.next_request_id() - 1
    )
    await con[1].send(msg.make_response(RpcInvalidParamError("text")))
    with pytest.raises(RpcInvalidParamError, match=r"text"):
        await task


async def test_call_delay(con):
    """Check that we can call delayed requests."""
    reported = []
    task = asyncio.create_task(
        con[0].call("test", "delay", query_timeout=0.01, progress=reported.append)
    )
    msg = await con[1].receive()
    assert msg == RpcMessage.request(
        "test", "delay", None, RpcMessage.next_request_id() - 1
    )
    await con[1].send(msg.make_response_delay(0.2))
    await con[1].send(msg.make_response_delay(0.3))
    assert await con[1].receive() == msg.make_abort(False)
    await con[1].send(msg.make_response())
    assert await task is None
    assert reported == [0.2, 0.3]


async def test_call_lost(con):
    """Check that we send request again as soon as lost is detected."""
    task = asyncio.create_task(con[0].call("test", "lost", query_timeout=0.01))
    msg = await con[1].receive()
    rid = RpcMessage.next_request_id() - 1
    assert msg == RpcMessage.request("test", "lost", None, rid)
    await con[1].send(msg.make_response(RpcRequestInvalidError("Lost")))
    assert await con[1].receive() == RpcMessage.request("test", "lost", None, rid)
    await con[1].send(msg.make_response("value"))
    assert await task == "value"


async def test_call_reset(con):
    """Check that we will recover call after reset."""
    task = asyncio.create_task(con[0].call("test", "withreset"))
    msg = await con[1].receive()
    rid = RpcMessage.next_request_id() - 1
    assert msg == RpcMessage.request("test", "withreset", None, rid)
    await con[1].reset()
    msg = await con[1].receive()
    assert msg == RpcMessage.request("test", "withreset", None, rid)
    await con[1].send(msg.make_response(42))
    assert await task == 42


async def test_call_user_id_required(con):
    """Check that we are able to fill in the user ID."""
    con[0].user_id = "foo"
    task = asyncio.create_task(con[0].call("test", "foruserid"))
    msg = await con[1].receive()
    assert msg == RpcMessage.request(
        "test", "foruserid", None, RpcMessage.next_request_id() - 1
    )
    await con[1].send(msg.make_response(RpcUserIDRequiredError()))
    msg = await con[1].receive()
    assert msg == RpcMessage.request(
        "test", "foruserid", None, RpcMessage.next_request_id() - 1, user_id="foo"
    )
    await con[1].send(msg.make_response(msg.user_id))
    assert await task == "foo"


async def test_call_try_again_later(con):
    """Check that we are able to delay our request after try again later error."""
    task = asyncio.create_task(con[0].call("test", "again", retry_timeout=0.05))
    msg = await con[1].receive()
    assert msg == RpcMessage.request(
        "test", "again", None, RpcMessage.next_request_id() - 1
    )
    await con[1].send(msg.make_response(RpcTryAgainLaterError()))
    msg = await con[1].receive()
    assert msg == RpcMessage.request(
        "test", "again", None, RpcMessage.next_request_id() - 1
    )
    await con[1].send(msg.make_response(42))
    assert await task == 42


async def test_call_abort(con):
    """Check that we are able to delay our request after try again later error."""
    task = asyncio.create_task(con[0].call("test", "abort"))
    msg = await con[1].receive()
    assert msg == RpcMessage.request(
        "test", "abort", None, RpcMessage.next_request_id() - 1
    )
    task.cancel()
    assert await con[1].receive() == msg.make_abort(True)
    with pytest.raises(asyncio.CancelledError):
        await task


async def test_call_old(con):
    """Check that we can manage old calls if message is lost."""
    task = asyncio.create_task(con[0].call("test", "old", retry_timeout=0.05))
    msg = await con[1].receive()
    rid = RpcMessage.next_request_id() - 1
    assert msg == RpcMessage.request("test", "old", None, rid)
    assert await con[1].receive() == msg
    await con[1].send(msg.make_response(0))
    assert await task == 0



"""Logger wrapper that overwrites debug method used for can package."""

import logging


def patchlogger(logger: logging.Logger) -> None:
    setattr(logger, "debug", lambda *args, **kwargs: logger.log(9, *args, **kwargs))  # noqa: B010
    setattr(logger, "info", lambda *args, **kwargs: logger.log(9, *args, **kwargs))  # noqa: B010



"""Check the RPC Types abstraction."""

import datetime
import decimal
import re

import pytest

from shv.rpctypes import (
    RpcTypeAny,
    RpcTypeBitfield,
    RpcTypeBlob,
    RpcTypeBool,
    RpcTypeDateTime,
    RpcTypeDecimal,
    RpcTypeDouble,
    RpcTypeEnum,
    RpcTypeIMap,
    RpcTypeInteger,
    RpcTypeKeyStruct,
    RpcTypeList,
    RpcTypeMap,
    RpcTypeNull,
    RpcTypeOneOf,
    RpcTypeParseError,
    RpcTypeString,
    RpcTypeStruct,
    RpcTypeTuple,
    RpcTypeUnsigned,
    rpctype_alert,
    rpctype_any,
    rpctype_blob,
    rpctype_bool,
    rpctype_clientinfo,
    rpctype_datetime,
    rpctype_decimal,
    rpctype_dir,
    rpctype_double,
    rpctype_exchange_p,
    rpctype_exchange_r,
    rpctype_exchange_v,
    rpctype_getlog_p,
    rpctype_getlog_r,
    rpctype_history_records,
    rpctype_imap,
    rpctype_integer,
    rpctype_list,
    rpctype_map,
    rpctype_null,
    rpctype_parse,
    rpctype_stat,
    rpctype_string,
    rpctype_unsigned,
)

DATA = [
    ("n", rpctype_null),
    ("b", rpctype_bool),
    ("i", RpcTypeInteger()),
    ("ikG", RpcTypeInteger(unit="kG")),
    ("i(-2,2)", RpcTypeInteger(-2, 2)),
    ("i(,2)", RpcTypeInteger(None, 2)),
    ("i(-2,)%", RpcTypeInteger(-2, None, "%")),
    ("i(^7,>8)", RpcTypeInteger(128, 255)),
    ("u", RpcTypeUnsigned()),
    ("uK", RpcTypeUnsigned(unit="K")),
    ("u(2)", RpcTypeUnsigned(0, 2)),
    ("u(1,10)", RpcTypeUnsigned(1, 10)),
    ("u(125,)%", RpcTypeUnsigned(125, None, "%")),
    ("u(^7,>8)", RpcTypeUnsigned(128, 255)),
    ("i[FALSE,TRUE]", RpcTypeEnum({0: "FALSE", 1: "TRUE"})),
    ("i[fail:-1,success]", RpcTypeEnum({-1: "fail", 0: "success"})),
    ("f", RpcTypeDouble()),
    ("fkG", RpcTypeDouble("kG")),
    ("d", RpcTypeDecimal()),
    ("d(-.3,.8)", RpcTypeDecimal(decimal.Decimal("-0.3"), decimal.Decimal("0.8"))),
    (
        "d(0,100,2)%",
        RpcTypeDecimal(decimal.Decimal("0"), decimal.Decimal("100"), 2, "%"),
    ),
    ("d(,,-2)", RpcTypeDecimal(None, None, -2)),
    ("s", RpcTypeString()),
    ("s(16)", RpcTypeString(16, 16)),
    ("s(,63)", RpcTypeString(0, 63)),
    ("x", RpcTypeBlob()),
    ("x(16)", RpcTypeBlob(16, 16)),
    ("x(,63)", RpcTypeBlob(0, 63)),
    ("t", rpctype_datetime),
    ("[i(0,100)]", RpcTypeList(RpcTypeInteger(0, 100))),
    ("[?](8)", RpcTypeList(rpctype_any, 8, 8)),
    ("[?](1,4)", RpcTypeList(rpctype_any, 1, 4)),
    ("[?](8,)", RpcTypeList(rpctype_any, 8)),
    ("[?](,4)", RpcTypeList(rpctype_any, 0, 4)),
    ("[d:x,d:y]", RpcTypeTuple((RpcTypeDecimal(), "x"), (RpcTypeDecimal(), "y"))),
    ("i{?}", RpcTypeIMap(rpctype_any)),
    ("i{i(3,9)}", RpcTypeIMap(RpcTypeInteger(3, 9))),
    (
        "i{t:date,i(0,63):level,s:id,?:info}",
        RpcTypeStruct({
            0: (rpctype_datetime, "date"),
            1: (RpcTypeInteger(0, 63), "level"),
            2: (rpctype_string, "id"),
            3: (rpctype_any, "info"),
        }),
    ),
    (
        "i{s:name:1,t:birth}",
        RpcTypeStruct({1: (rpctype_string, "name"), 2: (rpctype_datetime, "birth")}),
    ),
    ("{?}", RpcTypeMap(rpctype_any)),
    ("{i(0,2)}", RpcTypeMap(RpcTypeInteger(0, 2))),
    (
        "{s:name,t:birth}",
        RpcTypeKeyStruct({"name": rpctype_string, "birth": rpctype_datetime}),
    ),
    (
        "u[b:bool,u(3):umax:4,u(2,3):ushift,i[one,two]:enum]",
        RpcTypeBitfield(
            (0, rpctype_bool, "bool"),
            (4, RpcTypeUnsigned(0, 3), "umax"),
            (6, RpcTypeUnsigned(2, 3), "ushift"),
            (7, RpcTypeEnum({0: "one", 1: "two"}), "enum"),
        ),
    ),
    ("?", rpctype_any),
    ("?(Some type)", RpcTypeAny("Some type")),
    ("i|n", RpcTypeOneOf(rpctype_integer, rpctype_null)),
    ("i|s|d", RpcTypeOneOf(rpctype_integer, rpctype_string, rpctype_decimal)),
    ("!dir", rpctype_dir),
    ("!alert", rpctype_alert),
    ("!clientInfo", rpctype_clientinfo),
    ("!stat", rpctype_stat),
    ("!exchangeP", rpctype_exchange_p),
    ("!exchangeR", rpctype_exchange_r),
    ("!exchangeV", rpctype_exchange_v),
    ("!getLogP", rpctype_getlog_p),
    ("!getLogR", rpctype_getlog_r),
    ("!historyRecords", rpctype_history_records),
]


@pytest.mark.parametrize(
    "text,obj",
    (
        *DATA,
        ("i(,)", RpcTypeInteger()),
        ("i(0,)", RpcTypeInteger(0)),
        ("u(,2)", RpcTypeUnsigned(0, 2)),
        ("u(,0)", RpcTypeUnsigned(0, 0)),
    ),
)
def test_text2obj(text, obj):
    """Check that text can be converted to the object and is equal."""
    assert rpctype_parse(text) == obj


@pytest.mark.parametrize("text,obj", DATA)
def test_obj2text(text, obj):
    """Check that object can be converted to the object and is the same."""
    assert str(obj) == text


@pytest.mark.parametrize(
    "text,errmsg",
    (
        ("", "Missing character at 0"),
        ("i(", "Expected ',' at 2"),
        ("i[val:text]", "Expected number at 6"),
        ("i[:1]", "Text expected at 2"),
        ("u[i:item]", "Type 'i' can't be in Bitfield"),
        ("u[u(4,):item]", "Type 'u(4,)' can't be in Bitfield"),
        ("u[u(7):one,u(3):two:2]", "Bit 2 is used in multiple items"),
        ("i[again,again]", "Duplicates are not allowed in enum"),
        ("i[zero,again:0]", "Reuse of the integer key 0"),
        ("[i]Nothing", "Unexpected character 'N' at 3"),
        ("i{i:same,i:same}", "Duplicate keys are not allowed"),
        ("i{i:one,i:two:0}", "Reuse of the integer key 0"),
        ("{i:same,i:same}", "Reuse of the key 'same'"),
        ("[s](2,1)", "Minimum is greater than maximum"),
        ("s(2,1)", "Minimum is greater than maximum"),
        ("x(2,1)", "Minimum is greater than maximum"),
        ("d(2,1)", "Minimum is greater than maximum"),
        ("i(2,1)", "Minimum is greater than maximum"),
        ("u(2,1)", "Minimum is greater than maximum"),
        ("!invalid", "Invalid standard type 'invalid'"),
    ),
)
def test_textfail(text, errmsg):
    """Check the error message when parsing the RPC type definition."""
    with pytest.raises(RpcTypeParseError, match=f"^{re.escape(errmsg)}$"):
        rpctype_parse(text)


@pytest.mark.parametrize(
    "obj,value",
    (
        (rpctype_null, None),
        (rpctype_bool, True),
        (rpctype_bool, False),
        (rpctype_integer, 42),
        (rpctype_unsigned, 42),
        (rpctype_datetime, datetime.datetime.fromtimestamp(0)),
        (rpctype_decimal, decimal.Decimal("1.0")),
        (rpctype_double, 4.2),
        (rpctype_string, "foo"),
        (rpctype_blob, b"foo"),
        (rpctype_list, [42, "foo"]),
        (rpctype_map, {"foo": 42}),
        (rpctype_imap, {42: "foo"}),
        (rpctype_any, None),
        (RpcTypeEnum({1: "foo", 3: "bar"}), 1),
        (
            RpcTypeTuple(
                (rpctype_bool, "one"),
                (RpcTypeOneOf(rpctype_bool, rpctype_null), "second"),
            ),
            [True],
        ),
        (
            RpcTypeTuple(
                (rpctype_bool, "one"),
                (RpcTypeOneOf(rpctype_bool, rpctype_null), "second"),
            ),
            [True, True],
        ),
        (
            RpcTypeStruct({
                1: (rpctype_string, "name"),
                2: (RpcTypeOneOf(rpctype_datetime, rpctype_null), "birth"),
            }),
            {1: "john", 2: datetime.datetime.fromtimestamp(42)},
        ),
        (
            RpcTypeStruct({
                1: (rpctype_string, "name"),
                2: (RpcTypeOneOf(rpctype_datetime, rpctype_null), "birth"),
            }),
            {1: "john"},
        ),
        (
            RpcTypeKeyStruct({
                "name": rpctype_string,
                "birth": RpcTypeOneOf(rpctype_datetime, rpctype_null),
            }),
            {"name": "john", "birth": datetime.datetime.fromtimestamp(42)},
        ),
        (
            RpcTypeKeyStruct({
                "name": rpctype_string,
                "birth": RpcTypeOneOf(rpctype_datetime, rpctype_null),
            }),
            {"name": "john"},
        ),
        (
            RpcTypeBitfield(
                (0, rpctype_bool, "bool"),
                (4, RpcTypeUnsigned(0, 3), "umax"),
                (6, RpcTypeUnsigned(2, 3), "ushift"),
                (7, RpcTypeEnum({0: "one", 1: "two"}), "enum"),
            ),
            0,
        ),
        (
            RpcTypeBitfield(
                (0, rpctype_bool, "bool"),
                (4, RpcTypeUnsigned(0, 3), "umax"),
                (6, RpcTypeUnsigned(2, 3), "ushift"),
                (7, RpcTypeEnum({0: "one", 1: "two"}), "enum"),
            ),
            1 + (2 << 4) + (1 << 6) + (1 << 7),
        ),
        (
            rpctype_alert,
            {0: datetime.datetime.fromtimestamp(44), 1: 42, 2: "TestAlert"},
        ),
    ),
)
def test_validate(obj, value):
    """Check validation of the valid types."""
    assert obj.validate(value)


@pytest.mark.parametrize(
    "obj,value",
    (
        (rpctype_null, 1),
        (rpctype_bool, None),
        (RpcTypeInteger(0, 3), 42),
        (rpctype_unsigned, -1),
        (rpctype_datetime, 42),
        (rpctype_decimal, 42),
        (rpctype_double, 42),
        (rpctype_string, b"foo"),
        (rpctype_blob, "foo"),
        (rpctype_list, 42),
        (rpctype_map, {42: "foo"}),
        (rpctype_imap, {"foo": 42}),
        (RpcTypeEnum({1: "foo", 3: "bar"}), 2),
        (RpcTypeOneOf(rpctype_bool, rpctype_null), 42),
        (
            RpcTypeTuple(
                (rpctype_bool, "one"),
                (RpcTypeOneOf(rpctype_bool, rpctype_null), "second"),
            ),
            [42],
        ),
        (
            RpcTypeTuple(
                (rpctype_bool, "one"),
                (RpcTypeOneOf(rpctype_bool, rpctype_null), "second"),
            ),
            42,
        ),
        (
            RpcTypeTuple(
                (rpctype_bool, "one"),
                (RpcTypeOneOf(rpctype_bool, rpctype_null), "second"),
            ),
            [True, True, True],
        ),
        (
            RpcTypeStruct({
                1: (rpctype_string, "name"),
                2: (RpcTypeOneOf(rpctype_datetime, rpctype_null), "birth"),
            }),
            {},
        ),
        (
            RpcTypeStruct({
                1: (rpctype_string, "name"),
                2: (RpcTypeOneOf(rpctype_datetime, rpctype_null), "birth"),
            }),
            {1: "john", 3: 42},
        ),
        (
            RpcTypeKeyStruct({
                "name": rpctype_string,
                "birth": RpcTypeOneOf(rpctype_datetime, rpctype_null),
            }),
            {},
        ),
        (
            RpcTypeKeyStruct({
                "name": rpctype_string,
                "birth": RpcTypeOneOf(rpctype_datetime, rpctype_null),
            }),
            {"name": "john", "invalid": 42},
        ),
        (
            RpcTypeBitfield(
                (0, rpctype_bool, "bool"),
                (4, RpcTypeUnsigned(0, 3), "umax"),
                (6, RpcTypeUnsigned(2, 3), "ushift"),
                (7, RpcTypeEnum({0: "one", 1: "two"}), "enum"),
            ),
            2,
        ),
    ),
)
def test_validate_invalid(obj, value):
    """Check validation of the invalid types."""
    assert not obj.validate(value)


@pytest.mark.parametrize(
    "obj1,obj2",
    (
        (rpctype_null, RpcTypeNull()),
        (rpctype_bool, RpcTypeBool()),
        (rpctype_integer, RpcTypeInteger()),
        (rpctype_unsigned, RpcTypeUnsigned()),
        (rpctype_datetime, RpcTypeDateTime()),
        (rpctype_double, RpcTypeDouble()),
        (rpctype_decimal, RpcTypeDecimal()),
        (rpctype_string, RpcTypeString()),
        (rpctype_blob, RpcTypeBlob()),
        (rpctype_list, RpcTypeList()),
        (rpctype_map, RpcTypeMap()),
        (rpctype_imap, RpcTypeIMap()),
    ),
)
def test_singletons(obj1, obj2):
    assert obj1 is obj2


def test_integer_attr():
    assert rpctype_integer.minimum is None
    assert rpctype_integer.maximum is None
    assert not rpctype_integer.unit
    with pytest.raises(ValueError, match=r"^Unit contains forbidden characters$"):
        RpcTypeInteger(unit="]")


def test_unsigned_attr():
    assert rpctype_unsigned.minimum == 0
    assert rpctype_unsigned.maximum is None
    assert not rpctype_unsigned.unit
    with pytest.raises(ValueError, match=r"^Unit contains forbidden characters$"):
        RpcTypeUnsigned(unit="]")


def test_decimal_attr():
    assert rpctype_decimal.minimum is None
    assert rpctype_decimal.maximum is None
    assert rpctype_decimal.precision is None
    assert not rpctype_decimal.unit
    with pytest.raises(ValueError, match=r"^Unit contains forbidden characters$"):
        RpcTypeDecimal(unit="]")


def test_double_attr():
    assert not rpctype_double.unit
    with pytest.raises(ValueError, match=r"^Unit contains forbidden characters$"):
        RpcTypeDouble(unit="]")


def test_string_attr():
    assert rpctype_string.minlen == 0
    assert rpctype_string.maxlen is None


def test_blob_attr():
    assert rpctype_blob.minlen == 0
    assert rpctype_blob.maxlen is None


def test_list_attr():
    assert rpctype_list.type is rpctype_any
    assert rpctype_list.minlen == 0
    assert rpctype_list.maxlen is None


def test_map_attr():
    assert rpctype_map.type is rpctype_any


def test_imap_attr():
    assert rpctype_imap.type is rpctype_any


def test_enum_attr():
    tp = RpcTypeEnum({1: "foo", 3: "bar"})
    assert len(tp) == 2


def test_tuple_attr():
    tp = RpcTypeTuple(
        (rpctype_bool, "one"), (RpcTypeOneOf(rpctype_bool, rpctype_null), "second")
    )
    assert len(tp) == 2
    assert tp[0] == (rpctype_bool, "one")
    assert tp[0:2] == (
        (rpctype_bool, "one"),
        (RpcTypeOneOf(rpctype_bool, rpctype_null), "second"),
    )


def test_struct_attr():
    tp = RpcTypeStruct({1: (rpctype_string, "name"), 2: (rpctype_datetime, "birth")})
    assert len(tp) == 2
    assert tp[1] == (rpctype_string, "name")
    assert tp.key("name") == 1
    assert tp.key("birth") == 2
    with pytest.raises(KeyError):
        tp.key("invalid")
    assert list(tp) == [1, 2]


def test_keystruct_attr():
    tp = RpcTypeKeyStruct({"name": rpctype_string, "birth": rpctype_datetime})
    assert len(tp) == 2
    assert tp["name"] is rpctype_string
    assert list(tp) == ["name", "birth"]


def test_bitfield_attr():
    tp = RpcTypeBitfield((0, rpctype_bool, "bool"), (4, RpcTypeUnsigned(0, 3), "umax"))
    assert len(tp) == 2
    assert list(tp) == [
        (0, rpctype_bool, "bool"),
        (4, RpcTypeUnsigned(0, 3), "umax"),
    ]


def test_any_attr():
    assert not rpctype_any.alias


@pytest.mark.parametrize(
    ("tp", "name", "text"),
    (
        (
            rpctype_dir,
            "dir",
            "i{s:name:1,u[b:isGetter:1,b:isSetter,b:largeResult,b:notIndempotent,b:userIDRequired,b:isUpdatable]|n:flags,s|n:paramType,s|n:resultType,i(0,63):accessLevel,{s|n}:signals,{?}:extra:63}",
        ),
        (rpctype_alert, "alert", "i{t:date,i(0,63):level,s:id,?:info}"),
        (
            rpctype_clientinfo,
            "clientInfo",
            "i{i:clientId:1,s|n:userName,s|n:mountPoint,{i|n}|n:subscriptions,{?}:extra:63}",
        ),
        (
            rpctype_stat,
            "stat",
            "i{i:type,i:size,i:pageSize,t|n:accessTime,t|n:modTime,i|n:maxWrite}",
        ),
        (rpctype_exchange_p, "exchangeP", "i{u:counter,u|n:readyToReceive,x|n:data:3}"),
        (
            rpctype_exchange_r,
            "exchangeR",
            "i{u|n:readyToReceive:1,u|n:readyToSend,x|n:data}",
        ),
        (rpctype_exchange_v, "exchangeV", "i{u|n:readyToReceive:1,u|n:readyToSend}"),
        (
            rpctype_getlog_p,
            "getLogP",
            "i{t|n:since:1,t|n:until,i(0,)|n:count,s|n:ri}",
        ),
        (
            rpctype_getlog_r,
            "getLogR",
            "[i{t|n:timestamp:1,i(0,)|n:ref,s|n:path,s|n:signal,s|n:source,?:value,s|n:userId,b|n:repeat}]",
        ),
        (
            rpctype_history_records,
            "historyRecords",
            "[i{i[normal:1,keep,timeJump,timeAbig]:type,t:timestamp,s|n:path,s|n:signal,s|n:source,?:value,i(0,63):accessLevel,s|n:userId,b|n:repeat,i|n:timeJump:60}]",
        ),
    ),
)
def test_standard_repr(tp, name, text):
    """Compare standard types against representation in standard."""
    assert tp.name == name
    assert str(tp) == f"!{name}"
    assert str(tp.type) == text



"""The SHV RPC type list."""

from __future__ import annotations

import collections.abc
import typing

from .. import SHVType
from .any import rpctype_any
from .base import RpcType


class RpcTypeList(RpcType):
    """The List type representation."""

    __obj: RpcTypeList | None = None

    def __new__(  # noqa: D102
        cls, tp: RpcType = rpctype_any, minlen: int = 0, maxlen: int | None = None
    ) -> RpcTypeList:
        if tp is rpctype_any and minlen == 0 and maxlen is None:
            if cls.__obj is None:
                cls.__obj = super().__new__(cls)
            return cls.__obj
        return super().__new__(cls)

    def __init__(
        self, tp: RpcType = rpctype_any, minlen: int = 0, maxlen: int | None = None
    ) -> None:
        if maxlen is not None and minlen > maxlen:
            raise ValueError("Minimum is greater than maximum")
        if minlen < 0 or (maxlen is not None and maxlen < 0):
            raise ValueError("List size can be only positive number")
        self._tp = tp
        self._min = minlen
        self._max = maxlen

    @property
    def type(self) -> RpcType:
        """Content type allowed to be used in the list."""
        return self._tp

    @property
    def minlen(self) -> int:
        """Minimum number of items in the list."""
        return self._min

    @property
    def maxlen(self) -> int | None:
        """Maximum number of items in the list."""
        return self._max

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, RpcTypeList)
            and self._tp == other._tp
            and self._min == other._min
            and self._max == other._max
        )

    def __str__(self) -> str:
        lim = ""
        if self._min != 0 or self._max is not None:
            if self._min == self._max:
                lim = f"({self._max})"
            else:
                lim = f"({self._min or ''},{self._max or ''})"
        return f"[{self._tp}]{lim}"

    def validate(self, value: SHVType) -> typing.TypeGuard[collections.abc.Sequence]:  # noqa: D102
        return isinstance(value, collections.abc.Sequence) and all(
            self._tp.validate(v) for v in value
        )


rpctype_list = RpcTypeList()



"""The SHV RPC type Unsigned integer."""

from __future__ import annotations

import typing

from .. import SHVType
from ._tools import strnum as _strnum
from .base import RpcType


class RpcTypeUnsigned(RpcType):
    """The Unsigned type representation."""

    __obj: RpcTypeUnsigned | None = None

    def __new__(  # noqa: D102
        cls,
        minimum: int = 0,
        maximum: int | None = None,
        unit: str = "",
    ) -> RpcTypeUnsigned:
        if minimum == 0 and maximum is None and not unit:
            if cls.__obj is None:
                cls.__obj = super().__new__(cls)
            return cls.__obj
        return super().__new__(cls)

    def __init__(
        self,
        minimum: int = 0,
        maximum: int | None = None,
        unit: str = "",
    ) -> None:
        if maximum is not None and minimum > maximum:
            raise ValueError("Minimum is greater than maximum")
        if minimum < 0 or (maximum is not None and maximum < 0):
            raise ValueError("Unsigned limit can't be negative number")
        self._min = minimum
        self._max = maximum
        if any(c in "[]{}():,|" for c in unit):
            raise ValueError("Unit contains forbidden characters")
        self._unit = unit

    @property
    def minimum(self) -> int:
        """Minimum value allowed for this Int."""
        return self._min

    @property
    def maximum(self) -> int | None:
        """Maximum value allowed for this Int."""
        return self._max

    @property
    def unit(self) -> str:
        """Unit to be appended to the number."""
        return self._unit

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, RpcTypeUnsigned)
            and self._min == other._min
            and self._max == other._max
            and self._unit == other._unit
        )

    def __str__(self) -> str:
        lim = ""
        if self._min != 0 or self._max is not None:
            if self._min == 0:
                lim = f"({_strnum(self._max)})"
            else:
                lim = f"({_strnum(self._min)},{_strnum(self._max)})"
        return f"u{lim}{self._unit}"

    def validate(self, value: SHVType) -> typing.TypeGuard[int]:  # noqa: D102
        return (
            isinstance(value, int)
            and value >= 0
            and (self._min is None or value >= self._min)
            and (self._max is None or value <= self._max)
        )


rpctype_unsigned = RpcTypeUnsigned()
